/****
*
*      oFont.prg
*      oObject Class Demonstration.
*
*      Notes:
*      ------
*      VGA Soft Font loader/unloader. Uses standard Soft font files.
*      The most usual are 4096 byte files (16x8 font cell).
*
*      This class implementation also loads 8x8 and 14x8 fonts. The
*      file 28.fon contains a 14x8 font that - when loaded - produces
*      a 28 line display.
*
*      The ::ROM*() methods load the default Rom fonts found in every
*      VGA Rom Bios.
*
*      Note(1): Makes heavy use of FT_INT86() found in Nanforum Toolkit.
*
*      Note(2): It can easily inherit the oInt class found in oInt.prg in
*               this directory.
*
*/

#include "oObject.ch"
#include "ftint86.ch"

#define  x10        16
#define  x11        17
#define  x12        18
#define  x14        20
#define INT_VIDEO   16

#ifdef TEST

Proc TestFont()
Local oFont := Font():New( "modern.fon" )
   oFont:Load():Set()
   ? "Using 'Modern.Fon' CharSet..."
   ? "New Font Set - 25 MaxLines..."
   inkey(0)

   oFont:Reset()
   oFont:FontFile := "28.fon"
   oFont:Load():Set()
   ?
   ? "Using '28.Fon' CharSet..."
   ? "New Font Set - 28 MaxLines..."
   inkey(0)

   ?
   ? "Reset Font to default 8x8 - 50 MaxLines..."
   oFont:Set8x8()
   inkey(0)

   ?
   ? "Reset Font to default 14x8 (B/W) - 28 MaxLines..."
   oFont:Reset()
   inkey(0)

   ?
   ? "Reset Font to default 16x8 - 25 MaxLines..."
   oFont:Reset()
   oFont:Kill()

RETURN
#endif

//

CLASS Font ALIAS oFnt

DATA lIsSet          AS .F.
DATA FontFile        AS ""    // Soft-Font Filename to use
DATA FontBuffer      AS ""    // The buffer in memory containing the soft font
DATA nCharHeight     AS 0     // Height of each char. (BufferSize/256)
DATA nCharWidth      AS 8     // Width of each char. (default 8)

METHOD Load()        INLINE ::FontBuffer  := MemoRead( ::FontFile ),;
                            ::nCharHeight := LEN(::FontBuffer)/256

METHOD New( cFile )  INLINE ::FontFile   := IF( cFile == NIL, "", cFile )

METHOD Set()
METHOD Reset()       INLINE ::Set8x16()
METHOD ROM8x8()
METHOD ROM8x14()
METHOD ROM8x16()
METHOD Kill()

ENDCLASS

//
// Loads soft fonts of any height into VGA buffer
//
METHOD Set()
//
Local aRegs[10]

IF LEN( ::FontBuffer ) <= 0
   RETURN self
ENDIF

::lIsSet      := .T.
::nCharHeight := Len( ::FontBuffer ) / 256   // -> 256 chars per file

aRegs[AX] := MAKEHI(x11)+ x10
aRegs[BX] := MAKEHI( ::nCharHeight )+0
aRegs[CX] := 256
aRegs[DX] := 0
aRegs[ES] := ::FontBuffer+" "
aRegs[BP] := REG_ES
FT_INT86( 16, aRegs )

RETURN( self )


//
// Load default 8x14 Font - only on B/W Monitors
//
METHOD ROM8x14()
Local aRegs[10]
aRegs[AX] := MAKEHI(x11)+ x11
aRegs[BX] := 0
FT_INT86( 16, aRegs )
::lIsSet  := .F.
RETURN self


// Load default 8x8 Font
//
METHOD Set8x8()
Local aRegs[10]
aRegs[AX] := MAKEHI(x11)+ x12
aRegs[BX] := 0
FT_INT86( 16, aRegs )
::lIsSet  := .F.
RETURN self


// Load default 8x16 Font
//
METHOD Set8x16()
Local aRegs[10]
aRegs[AX] := MAKEHI(x11)+ x14
aRegs[BX] := 0
FT_INT86( 16, aRegs )
::lIsSet  := .F.
RETURN self


//
METHOD Kill()
::Reset()
::FontFile    := ""
::FontBuffer  := ""
::nCharHeight := 0
::lIsSet      := .F.
RETURN self


