#include "oobject.ch"

#translate GETSET( <xVal>, <xParm> ) => ;
           <xVal> := IF( <xParm> == NIL, <xVal>, <xVal> := <xParm>)


/****
*
*     Replacement for the TBrowse class #1.
*     Can be inherited using the oobject.lib
*
*     Note:
*     -----
*     This code is 5.3 specific. To convert it to 5.2 erase the 5.3 methods
*     and IVars from the class.
*
*/


#ifdef TEST

Proc test(cFile)
Local nKey,i
local oDb
     IF cFile == NIL .or. !file( cFile )
        ? "ERROR: NO VALID FILENAME GIVEN..."
        Quit
     endif

     USE (cFile) NEW
     oDb := DBrowseDB(2,2,20,75)

     oDb:ColorSpec := "n/w,w/b"
     oDb:Freeze    := 1
     oDb:headsep   := ""
     oDb:ColSep    := "  "

     for i=1 to fcount()
           oDb:AddColumn( oTbColumn():TbColumnNew( FieldName(i), FieldBlock( FieldName(i) ) ))
     next

     ? oDb:Freeze
     ? oDb:Autolite
     inkey(0)
     oDb:Setkey( 13, {|| alert( "Hello!!!") })
     oDb:ForceStable()

     WHILE (.T.)
       oDB:ForceStable()
           IF (oDB:ApplyKey(Inkey(0)) == -1)
       oDB:Hide()
       EXIT
     ENDIF
     ENDDO
RETURN

#endif


//------------------------------------------------------  SHELL FUNCTIONS
FUNCTION DBrowseNew(t,l,b,r)
Local o
DEFAULT t := 0, l := 0, b := MaxRow(), r := MaxCol()
o := oTBrowse():TBrowseNEW(t,l,b,r)
Return(o)

FUNCTION DBrowseDB(t,l,b,r)
Local o
DEFAULT t := 0, l := 0, b := MaxRow(), r := MaxCol()
o := oTBrowse():TBrowseDB(t,l,b,r)
Return(o)


/*************************************************************************
 This Class forms a container for tBrowse.  With this
 container class you can now 'inherit' from tbrowse. This
 class contains the tbrowse class with methods that map
 directly onto those of tbrowse.
 Note the set of 'assignment and retrievel messages
 for the tbrowse instance variables
 *************************************************************************/

//------------------------------------------------------  INIT CLASS
CLASS oTBrowse

DATA oTb
DATA cBack
DATA aArray, nIndex     // for array browsing
DATA oVScroll           // Vertical   ScrollBar obj
DATA oHScroll           // Horizontal ScrollBar obj

METHOD NEW             = _DBNEW
METHOD TBrowseDB       = _DBrowseDB
METHOD TBrowseNew      = _DBrowseNew
METHOD autoLite        = _autoLite
METHOD border          = _border
METHOD cargo           = _cargo
METHOD colCount        = _colCount
METHOD colorSpec       = _colorSpec
METHOD colPos          = _colPos
METHOD colSep          = _colSep
METHOD footSep         = _footSep
METHOD freeze          = _freeze
METHOD goBottomBlock   = _goBottomBlock
METHOD goTopBlock      = _goTopBlock
METHOD headSep         = _headSep
METHOD hitBottom       = _hitBottom
METHOD hitTop          = _hitTop
METHOD leftVisible     = _leftVisible
METHOD mColPos         = _mColPos          // 5.3
METHOD message         = _message          // 5.3
METHOD mRowPos         = _mRowPos          // 5.3
METHOD nBottom         = _nBottom
METHOD nLeft           = _nLeft
METHOD nRight          = _nRight
METHOD nTop            = _nTop
METHOD rightVisible    = _rightVisible
METHOD rowCount        = _rowCount
METHOD rowPos          = _rowPos
METHOD skipBlock       = _skipBlock
METHOD stable          = _stable

METHOD _autoLite       = _autoLite
METHOD _border         = _border
METHOD _cargo          = _cargo
METHOD _colCount       = _colCount
METHOD _colorSpec      = _colorSpec
METHOD _colPos         = _colPos
METHOD _colSep         = _colSep
METHOD _footSep        = _footSep
METHOD _freeze         = _freeze
METHOD _goBottomBlock  = _goBottomBlock
METHOD _goTopBlock     = _goTopBlock
METHOD _headSep        = _headSep
METHOD _hitBottom      = _hitBottom
METHOD _hitTop         = _hitTop
METHOD _leftVisible    = _leftVisible
METHOD _mColPos        = _mColPos         // 5.3
METHOD _message        = _message         // 5.3
METHOD _mRowPos        = _mRowPos         // 5.3
METHOD _nBottom        = _nBottom
METHOD _nLeft          = _nLeft
METHOD _nRight         = _nRight
METHOD _nTop           = _nTop
METHOD _rightVisible   = _rightVisible
METHOD _rowCount       = _rowCount
METHOD _rowPos         = _rowPos
METHOD _skipBlock      = _skipBlock
METHOD _stable         = _stable

METHOD down           = _down
METHOD end            = _end
METHOD goBottom       = _goBottom
METHOD goTop          = _goTop
METHOD home           = _home
METHOD left           = _left
METHOD pageDown       = _pageDown
METHOD pageUp         = _pageUp
METHOD panEnd         = _panEnd
METHOD panHome        = _panHome
METHOD panLeft        = _panLeft
METHOD panRight       = _panRight
METHOD right          = _right
METHOD up             = _up

METHOD addColumn      = _addColumn
METHOD applyKey       = _applyKey         // 5.3
METHOD colorRect      = _colorRect
METHOD colWidth       = _colWidth
METHOD configure      = _configure
METHOD deHilite       = _deHilite
METHOD delColumn      = _delColumn
METHOD forceStable    = _forceStable      // 5.3
METHOD getColumn      = _getColumn
METHOD hilite         = _hilite
METHOD hitTest        = _hitTest          // 5.3
METHOD insColumn      = _insColumn
METHOD invalidate     = _invalidate
METHOD refreshAll     = _refreshAll
METHOD refreshCurrent = _refreshCurrent
METHOD setColumn      = _setColumn
METHOD setKey         = _setKey           // 5.3
METHOD setstyle       = _setstyle         // 5.3
METHOD stabilize      = _stabilize

//--- new methods ---//
METHOD Hide           = __oTBhide
METHOD Display        = __oTBDisplay

ENDCLASS


//------------------------------------------------------  CONSTRUCTORS
METHOD FUNCTION _DBNew(t,l,b,r)
 ::oTb:= TBrowseDB(t,l,b,r)
 ::cBack := SaveScreen(t,l,b+1,r+1)
Return( self )

METHOD FUNCTION _DBrowseDB(t,l,b,r)
 ::oTb:= TBrowseDB(t,l,b,r)
 ::cBack := SaveScreen(t,l,b+1,r+1)
Return( self )

METHOD FUNCTION _DBrowseNew(t,l,b,r)
 ::oTb:= TBrowseNew(t,l,b,r)
 ::cBack := SaveScreen(t,l,b+1,r+1)
Return( self )


//------------------------------------------------------  DATA
FUNCTION _autoLite(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:Autolite, x ))

FUNCTION _border(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:border, x ))

FUNCTION _cargo(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:cargo, x ))

FUNCTION _colCount(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:colCount, x ))

FUNCTION _colorSpec(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:colorSpec, x ))

FUNCTION _colPos(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:colPos, x ))

FUNCTION _colSep(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:colSep, x ))

FUNCTION _footSep(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:footSep, x ))

FUNCTION _freeze(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:freeze, x ))

FUNCTION _goBottomBlock(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:goBottomBlock , x ))

FUNCTION _goTopBlock(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:goTopBlock, x ))

FUNCTION _headSep(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:headSep, x ))

FUNCTION _hitBottom(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:hitBottom, x ))

FUNCTION _hitTop(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:hitTop, x  ))

FUNCTION _leftVisible(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:leftVisible, x  ))

FUNCTION _mColPos(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:mColPos, x ))

FUNCTION _message(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:message, x ))


FUNCTION _mRowPos(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:mRowPos, x ))


FUNCTION _nBottom(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:nBottom, x ))


FUNCTION _nLeft(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:nLeft, x ))


FUNCTION _nRight(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:nRight, x ))


FUNCTION _nTop(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:nTop, x ))


FUNCTION _rightVisible(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:rightVisible, x  ))


FUNCTION _rowCount(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:rowCount, x  ))


FUNCTION _rowPos(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:rowPos, x  ))


FUNCTION _skipBlock(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:skipBlock, x ))


FUNCTION _stable(x)
Local Self := QSelf()
RETURN( GETSET( ::oTb:stable, x  ))


//------------------------------------------------------  METHODS
METHOD FUNCTION _down()
Return(DispBegin(),::oTb:down(), DispEnd())

METHOD FUNCTION _end()
Return(DispBegin(),::oTb:end(), DispEnd())

METHOD FUNCTION _goBottom()
Return(DispBegin(),::oTb:goBottom(), DispEnd())

METHOD FUNCTION _goTop()
Return(DispBegin(),::oTb:goTop(), DispEnd())

METHOD FUNCTION _home()
Return(DispBegin(),::oTb:home(), DispEnd())

METHOD FUNCTION _left()
Return(DispBegin(),::oTb:left(), DispEnd())

METHOD FUNCTION _pageDown()
Return(DispBegin(),::oTb:pageDown(), DispEnd())

METHOD FUNCTION _pageUp()
Return(DispBegin(),::oTb:pageUp(), DispEnd())

METHOD FUNCTION _panEnd()
Return(DispBegin(),::oTb:panEnd(), DispEnd())

METHOD FUNCTION _panHome()
Return(DispBegin(),::oTb:panHome(), DispEnd())

METHOD FUNCTION _panLeft()
Return(DispBegin(),::oTb:panLeft(), DispEnd())

METHOD FUNCTION _panRight()
Return(DispBegin(),::oTb:panRight(), DispEnd())

METHOD FUNCTION _right()
Return(DispBegin(),::oTb:right(), DispEnd())

METHOD FUNCTION _up()
Return(DispBegin(),::oTb:up(), DispEnd())

//------------------------------------------------------  MISC METHODS

METHOD FUNCTION _addColumn(x,y,z)
Return(::oTb:addColumn(x,y,z))

METHOD FUNCTION _applyKey(x)
Return(::oTb:applyKey(x))

METHOD FUNCTION _colorRect(x,y)
Return(::oTb:colorRect(x,y))

METHOD FUNCTION _colWidth(x)
Return(::oTb:colWidth(x))

METHOD FUNCTION _configure()
Return(::oTb:configure())

METHOD FUNCTION _deHilite()
Return(::oTb:deHilite())

METHOD FUNCTION _delColumn()
Return(::oTb:delColumn())

METHOD FUNCTION _forceStable()
Return(::oTb:forceStable())

METHOD FUNCTION _getColumn(x)
Return(::oTb:getColumn(x))

METHOD FUNCTION _hilite(x)
Return(::oTb:hilite(x))

METHOD FUNCTION _hitTest(x,y)
Return(::oTb:hitTest(x,y))

METHOD FUNCTION _insColumn(x,y,z)
Return(::oTb:insColumn(x,y,z))

METHOD FUNCTION _invalidate()
Return(::oTb:invalidate())

METHOD FUNCTION _refreshAll()
Return(::oTb:refreshAll())

METHOD FUNCTION _refreshCurrent()
Return(::oTb:refreshCurrent())

METHOD FUNCTION _setColumn(x,y,z)
Return(::oTb:setColumn(x,y,z))

METHOD FUNCTION _setKey(x,y,z)
Return(::oTb:setKey(x,y,z))

METHOD FUNCTION _setstyle(x)
Return(::oTb:setstyle(x))

METHOD FUNCTION _stabilize(x)
Return(::oTb:stabilize(x))


// NEW METHODS...
//------------------------------------------------------------------------

Method Function __oTBHide()
RestScreen(::ntop,::nleft,::nbottom+1,::nright+1, ::cBack)
RETURN (self)

METHOD Function __oTBDisplay()
DispBegin()
     ::oTB:RefreshAll():Forcestable()
DispEnd()
RETURN (self)
