/****
*     oRep.prg
*     oObject class demonstration.
*
*     NOTES:
*     ------
*     This is a flat database report object. It generates a text file
*     and is very flexible ( see the line counter example ).
*
*     It is not really oObject specific. You can easily convert it to
*     use any class engine you wish.
*
*     Read the oRep.NGT text and... Enjoy !!!
*/



#include "oRep.ch"

#define TEST          // undef to remove TestRep()

#ifdef TEST

*
PROC TestRep()
*
Local oRep
Local aHead
Local nOldRec
Local aData        := {}
Local nLineInc     := 0        // used to create a line counter...

USE ARRIV NEW
INDEX ON ARRIV->ARRIV TAG "ARRIV" TO ARRIV

Arriv->(DbGoTop())

aHead := { ;
          PadC("Arrivals List",132, " "),;
          PadC("",132, " ");
          }

oRep := Repo():New( aHead )

/*
   This will create a line list counter ( 1., 2., 3., ... ) .
   I convert it to string because it looks better... ;)
*/

oRep:LineBlock := {|| IF( Arriv->(EOF()), nLineInc := SPACE(3), nLineInc++)}

REPORT COLUMN "   "  ;
DATA {|| If( valtype(nLineInc) == "N", str(nLineInc,3), nLineInc ) }   ;
WIDTH 3              ;
In oRep

REPORT COLUMN "Code"  DATA {|| str(Arriv->Code,4) }   ;
WIDTH 4  ;
In oRep

REPORT COLUMN "Arrival"  DATA {|| Arriv->Arriv }   ;
WIDTH 8  ;
IN oRep

REPORT COLUMN "Depart."  DATA {|| Arriv->Depart};
WIDTH 8  ;
IN oRep

REPORT COLUMN "Name"    DATA {|| Arriv->name } ;
WIDTH 25 ;
In oRep

REPORT COLUMN "Pax"      DATA {|| Arriv->Pax } ;
PICTURE "9999" ;
WIDTH 4  ;
TOTAL    ;
In oRep

REPORT COLUMN "Tour Op." DATA {|| Arriv->Agent } ;
WIDTH 12 ;
In oRep

REPORT COLUMN "Hotel" DATA {|| Arriv->Hotel } ;
WIDTH 10 ;
In oRep


REPORT COLUMN " NET Total"     DATA {|| Arriv->Total_Net} ;
WIDTH 11 ;
TOTAL    ;
PICTURE "999,999,999" ;
In oRep

REPORT COLUMN "Commission"     DATA {|| Arriv->Total_Comi } ;
WIDTH 11 ;
TOTAL    ;
PICTURE "999,999,999" ;
In oRep

REPORT COLUMN " Total"    DATA {|| Arriv->Total_Face} ;
WIDTH 11 ;
TOTAL    ;
PICTURE "999,999,999" ;
In oRep

REPORT COLUMN " Sub."    DATA {|| Arriv->Total_Face - Arriv->Total_Net} ;
WIDTH 10 ;
TOTAL    ;
PICTURE "99,999,999" ;
In oRep

oRep:Cols           := 132
oRep:LinesPPage     := 30    // Cut it short. This is just a demo...
oRep:Colsep         := ""
oRep:HeadSep        := ""
oRep:FootSep        := ""
oRep:Pause          := .F.
oRep:lResetPage     := .F.
oRep:lTotalsPage    := .F.   // No separate totals page
oRep:lTotalEachPage := .T.
oRep:PrnInit        := ""
oRep:PrnEnd         := ""
oRep:EjectPage      := .T.
oRep:cInfo          := "ASCO Tours, Aegina - " + "  ("+DTOC( DATE() ) + ")"

oRep:Exec()
alert("All done ... Take a look at 'Report.txt'",,"w/b" )
scroll()

CLOSE DATABASES
Return

#endif



/*---------------------------------------------------------------------------*/
CLASS Repo ALIAS rep
/*---------------------------------------------------------------------------*/

EXPORT:

DATA   oWin                   // ... to display progress
DATA   Header
DATA   Footer
Data   Body
Data   ColSep
Data   HeadSep
Data   FootSep
Data   LinesPPage
Data   Cols
Data   PageNo
Data   EjectPage
Data   EjectLastPage
Data   bWhile         AS NIL
DATA   xSeek
Data   FHandle
Data   RepFile
DATA   cInfo                  // For ::InfoLine()
Data   PrnInit
DATA   PrnEnd
Data   Device                 // PRN
Data   Alias
Data   nRecCount
DATA   PageTotals             // Totals for each Page
DATA   lResetPage             // Reset Totals for each Page ???
DATA   Totals                 // Totals for the REPORT
DATA   lTotalsPage            // Generate Totals on Last Page ???
DATA   lTotalEachPage         // Generate Totals on Single Pages ???
DATA   Pause                  // Pause on every PageBreak
DATA   lKill
DATA   LineBlock
DATA   PageBlock

LOCAL:

DATA   LenHeader
DATA   LenFooter
DATA   LenBody
DATA   CurLine
DATA   RecsProcessed

EXPORT:

METHOD  New( aHeader, aFooter, aBody, cFile )
METHOD  AddColumn( aColumn )
METHOD  Exec()
METHOD  Kill( lKill )

LOCAL:

METHOD  MakeHeader()
METHOD  MakeFooter()
METHOD  MakeLine()
METHOD  MakePage()
METHOD  InfoLine()
METHOD  TotalsPage()

ENDCLASS



/*------------------------------------------------------------------*/
METHOD RepNew( aHeader, aFooter, aBody, cFile )

DEFAULT aHeader   := {""},;
        aFooter   := {""},;
        aBody     := {},;
        cFile     := "Report.Txt"

::Header          := aHeader
::Footer          := aFooter
::Body            := aBody
::HeadSep         := ""
::FootSep         := ""

::LenHeader       := Len( ::Header )
::LenFooter       := Len( ::Footer )
::PageNo          := 1
::EjectPage       :=.T.
::EjectLastPage   :=.T.
::lTotalsPage     :=.F.
::lTotalEachPage  :=.F.
::lResetPage      :=.F.
::nRecCount       := 0
::RecsProcessed   := 0
::RepFile         := cFile
::PrnInit         := ""
::PrnEnd          := ""
::LinesPPage      := 60
::HeadSep         := ""
::ColSep          := ""
::Cols            := 80
::LineBlock       := Nil
::PageBlock       := Nil
::CurLine         := 0
::cInfo := NIL

Return( Self )



/*------------------------------------------------------------------*/
METHOD Kill( lKill )

 ::lKill := lKill

Return( self )


/*------------------------------------------------------------------*/
METHOD AddColumn( aColumn )

DEFAULT aColumn[R_PIC]    := "@X"
DEFAULT aColumn[R_COLSEP] := NIL
DEFAULT aColumn[R_TOTAL]  := .F.

AADD( ::Body, aColumn )

Return( Self )



/*------------------------------------------------------------------*/
METHOD MakeHeader()
Local i

IF !Empty(::Header)
     For i = 1 to Len( ::Header )
          IF ::Header[i] != NIL .and. valtype( ::Header[i] ) == "C"
          FWriteLN( ::FHandle, ::Header[i] )
          ENDIF
     Next
ENDIF
Return( self )



/*------------------------------------------------------------------*/
METHOD Exec()
Local i, nKey := 0
Local lFound
Local nTmpRec

::FHandle         := FCreate( ::RepFile, FC_NORMAL )
IF Ferror() != 0
   ALERT("ERROR Creating "+ ::RepFile +"; ( " +_LAST_IO_ERROR +" )" )
   RETURN(SELF)
ENDIF

::LenHeader       := Len( ::Header )
::LenFooter       := Len( ::Footer )
::Totals          := Array( Len( ::Body ) )
::PageTotals      := Array( Len( ::Body ) )

Afill( ::Totals, 0 )
Afill( ::PageTotals, 0 )

  if ::bWhile == NIL .or. ::bWhile == ""
          ::bWhile := {|| .t. }
  endif

If ::xSeek != NIL
     lFound := DBSEEK( ::xSeek )
         If !lFound
               ::bWhile := {|| .t. }
               ::xSeek:= NIL
               DbGoTop()
         endif
else
     ::bWhile := {|| .t. }
     ::xSeek:= NIL
Endif

/* Count Recs */
nTmpRec  := Recno()
while !Eof()
    If Eval( ::bWhile ) .or. !Eof()
       ::nRecCount++
       DbSkip()
    Else
       Exit
    Endif
end
DbGoTo( nTmpRec )

/*
      // For progress display purpoces...
      DEFINE WINDOW ::oWin Height 5 WIDTH 50 CAPTION "Processing Report..."
*/

While !Eof()

   ::MakePage()
   ::PageNo ++

/* Reset Page Totals ??? */
IF ::lResetPage == .T.
     ASize( ::PageTotals, 0 )
     ::PageTotals:= Array( Len( ::Body ) )
       Afill( ::PageTotals, 0 )
Endif

/* -- Kill report if ESC pressed -- */
   If ::lKill == .t.
       Exit
   Endif

End

/* -- Build a Page for ::Totals -- */
IF ::lTotalsPage == .T.
     ::TotalsPage()
EndIf


/*
if !empty( ::PrnEnd )
    FWriteLN( ::FHandle, ::PrnEnd )
Endif
*/

/* -- Close Output File -- */
FWrite( ::FHandle, "", 1)
IF !FClose( ::FHandle )
    ALERT( "ERROR CLOSING FILE !!!; < "+_LAST_IO_ERROR+" > ")
ENDIF

/*
      // for progress display purpoces...
      DEACTIVATE WINDOW ::oWin
*/

Return( Self )



/*------------------------------------------------------------------*/
METHOD MakePage()
Local i,j:=1, nCount := 0
Local cLine    := ""    /* Report Line */
Local nKey     := 0
local nRec     := Recno()
Local cHeadStr := ""
Local cStr     := ""
Local xVal
Local nLenBody := Len( ::Body )
Local tmpBody, tmpColsep

/* Evaluate ::PageBlock */
IF ValType( ::PageBlock ) == "B"
      EVAL( ::PageBlock )
ENDIf


/*
  Reserve space for calc'ed fields
 5 = HeadLines + Sep + Footsep + InfoLine + Totals
*/
::LenBody := ::LinesPPage - ::LenHeader -::LenFooter - ;
             If(::lTotalEachPage == .t., 5, 4)


if !Empty( ::PrnInit ) .and. ::PageNo == 1
    FWriteLN( ::FHandle, ::PrnInit )
Endif

::MakeHeader()

    tmpBody   := ::Body
    tmpColsep := ::Colsep
    AEval( tmpBody, {|e| cHeadStr += Padr( e[1], ;
                                          e[3]  + ;
                                    If( !Empty(tmpColsep) .OR. ::Colsep!=NIL, Len(tmpColsep), Len(e[5])) )  } )
    FWriteLN( ::FHandle, cHeadStr )


    IF ::headSep != NIL .or. ::HeadSep != ""
         FWriteLN( ::FHandle, Replicate( ::HeadSep, ::Cols ) )
    Endif

WHILE j <= ::LenBody .and. !eof()

    if Eval( ::bWhile )
        ::MakeLine()
        nRec := Recno()
        ::RecsProcessed++

       /* // --->  Counter...
         // for progress display purpoces...
         @ ::oWin:nTop+2, ::oWin:nLeft+3 ;
             SAY "Page # " + padl( alltrim(lstr( ::PageNo )),3,"0") + space(3) +;
             ALLTRIM(lstr( ::nReccount )) + "/" + ALLTRIM(lstr( ::RecsProcessed ) ) ;
             COLOR ::oWin:Color
        */
        DbSkip()

        IF EOF()
           ::MakeLine()
        ENDIF

    ELSE

        ::MakeLine()
        EXIT
    ENDIF
        j++
ENDDO

    If ::Footsep != NIL .or. ::FootSep != ""
         FWriteLn( ::FHandle, Replicate( ::FootSep, ::Cols ) )
    Endif


/* --- Add Totals on each Page --- */
If ::lTotalEachPage

    cStr := ""
    FOR i = 1 to LEN( ::PageTotals )

        if ValType( ::Body[i, R_DATA] ) == "B"
            xVal := EVAL(::Body[i, R_DATA])
        Else
            xVal := FieldBlock(::Body[i, R_DATA])
        Endif

        If ValType( xVal ) == "N"
          cStr += PadL( transform(::PageTotals[i],::Body[i,R_PIC]), ::body[i,R_LEN] )
           if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
            Endif
        Else
          cStr += PadL( " ", ::Body[i,R_LEN] )
           // Last Column does not have a ColSep
           if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
           Endif
        Endif
    NEXT
    FWriteLn( ::FHandle, cStr )
Endif


::MakeFooter()
::InfoLine()

/* --- Printer Reset --- */
     IF Eof()
          if !empty( ::PrnEnd )
              FWriteLN( ::FHandle, ::PrnEnd )
          Endif
     EndIF

/*  --- Eject Page ??? --- */
If ( ::EjectPage , fwriteln(::FHandle, '' ), fwriteln(::FHandle, " ") )

If ::Pause == .T.
   nKey := Inkey(0)
        If nKey == K_ESC
            ::KillRepo(.T.)
        Endif
Endif
Return( self )


//
METHOD MakeLine()
//
Local nVal := 0
Local i, cStr := ""
Local nLenBody := Len(::Body)

IF Valtype( ::LineBlock ) == "B"
   Eval( ::LineBlock )
ENDIF

     FOR i = 1 to LEN( ::Body )
          IF Valtype( ::Body[i,R_DATA] ) == "B"

               nVal := Eval( ::Body[i,R_DATA] )

               /* CODE BLOCKS: No formatting ... */

               IF ValType( nVal ) == "N"
                    cStr += PadL( Transform(nVal, ::Body[i,R_PIC] ), ::Body[i,R_LEN] )
               ELSE
                    cStr += PadL( Transform(nVal, ::Body[i,R_PIC] ), ::Body[i,R_LEN] )
               EndIF

               if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
               Endif

                   IF ::Body[i, R_TOTAL] == .T. .and. ValType( nVal )== "N"
                   //Calculate Page totals
                       ::PageTotals[i] += nVal
                   //Calculate General totals
                       ::Totals[i] += nVal
                   EndIF

          ELSEIF IS_FIELD( ::Body[i,R_DATA] )
               nVal := Eval( FieldBlock( ::Body[ i, R_DATA ] ) )
               iF ValType( nVal ) == "N"
                    cStr +=  PadL( Transform( nVal, ::Body[i, R_PIC]), ::Body[i,R_LEN] )
               ELSE
                    cStr +=  PadL( Transform( nVal, ::Body[i, R_PIC]), ::Body[i,R_LEN] )
               Endif

               if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
               Endif

                   if ::Body[i,R_TOTAL] == .T. .and. ValType(nVal)== "N"
                       ::PageTotals[i] += nVal
                   Endif
          ENDIF
     NEXT i


     FWriteLN( ::FHandle , cStr )

return( self )


/*------------------------------------------------------------------*/
METHOD MakeFooter()
Local i
IF Len(::Footer) != 0
     For i = 1 to Len( ::Footer )
        FWriteLN( ::FHandle, ::Footer[i] )
     Next
ENDIF
return( self )


/*------------------------------------------------------------------*/
METHOD InfoLine()

FWriteLN( ::FHandle,  " Page (" + alltrim(lStr(::PageNo)) +") "  + ;
          if( ::cInfo!= NIL, ::cInfo, "" ) )

Return( self )


/*------------------------------------------------------------------*/
METHOD TotalsPage()
Local i, nKey  := 0
Local cHeadStr := ""
Local cStr     := ""
Local xVal
Local nLenBody := LEN( ::Body )
Local tmpBody, tmpColsep

    ::MakeHeader()

    tmpBody   := ::Body
    tmpColsep := ::Colsep
    AEval( tmpBody, {|e| cHeadStr += Padr( e[1], e[3]  + ;
                                    If( !Empty(tmpColsep) .OR. ::Colsep!=NIL, Len(tmpColsep), Len(e[5])) )  } )

         FWriteLN( ::FHandle, cHeadStr )

         If ::Headsep != NIL .or. ::HeadSep != ""
              FWriteLN( ::FHandle, Replicate( ::HeadSep, ::Cols ) )
         endif

         FOR i = 1 to Len( ::Totals )
             if valType( ::Body[i, R_DATA] ) == "B"
                 xVal := EVAL(::Body[i, R_DATA])
             Else
                 xVal := FieldBlock(::Body[i, R_DATA])
             Endif

             IF ValType(xVal ) == "N"
               cStr += PadL( Transform( ::Totals[i], ::Body[i,R_PIC] ), ::Body[i,R_LEN] )
                  if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
                  Endif

             ELSE
               cStr += PadL( " ", ::Body[i,R_LEN] )
                 if i < nLenBody
                   cStr += IF( ::body[i,R_COLSEP] !=NIL, ::Body[i,R_COLSEP], ::Colsep )
                 Endif
             Endif
         NEXT
               FWriteLn( ::FHandle, cStr )
     ::MakeFooter()
     ::InfoLine()


     /*  --- Eject Page ??? --- */
     If ( ::EjectPage , FWriteLn(::FHandle, '' ), FWriteLn( ::FHandle, " ") )

     If ::Pause == .T.
        nKey := Inkey(0)
     Endif

Return( Self )


/*---------------------------------------------------------------------------*/
STATIC FUNCTION FWriteLn( nHandle, cString, nLength, cDelim )
/*---------------------------------------------------------------------------*/
   IF cDelim == NIL
      cString += CHR(13) + CHR(10)
   ELSE
      cString += cDelim
   ENDIF

   RETURN ( FWRITE( nHandle, cString, nLength ) )


/* ---------------------------  END CLASS Repo() --------------------------- */


