#include "oobject.ch"
#include "box.ch"
#include "inkey.ch"
#include "manos.ch"
#include "o.ch"
#include "tbrowse.ch"

#translate GETSET( <xVal>, <xParm> ) => ;
           <xVal> := IF( <xParm> == NIL, <xVal>, <xVal> := <xParm>)

/****
*
*     Replacement for the TBrowse class # 2
*     Can be inherited using the oobject.lib
*
*     Note:
*     -----
*     This code is 5.3 specific. To convert it to 5.2 erase the 5.3 methods
*     and IVars from the class.
*
*     Description:
*     ------------
*     This all in one oTBrowse() Class incorporates 5.3 scroll bars,
*     mouse handling and combined Array - Table browsing ... and more
*
*     In version 2.B TBrowse and the rest native objects can be directly
*     inherited or SubClassed. This code however is a fraction faster than
*     Direct inheritance - Not subclassing.
*/


#ifdef TEST


//
Proc Test( cFile )
SetCursor(0)
MSetCursor(.t.)
SET EVENTMASK TO INKEY_ALL-INKEY_MOVE
alert(" ARRAY Browse Test... " )
scroll() ; TestArr()

alert(" TABLE Browse Test... " )
scroll() ; TestDB(cFile)
MSetCursor(.F.)
Return


//
Proc testArr()
Local oDB  := DBrowse():New( 10,10,20,70 )

         oDB:aArray    := Directory("*.*")
         oDb:ColorSpec := "n/w,w/b"
         oDb:headsep   := ""
         oDb:ColSep    := "  "

         oDB:aSetSkip()

         oDB:aAddColumn( "",  1, 20)
         oDB:aAddColumn( "",  2)
         oDB:aAddColumn( "",  3)
         oDB:aAddColumn( "",  4)

         oDb:Setkey( 13, {|| oDB:TbGet() })

         oDB:AddVSBar():AddHSBar()

         oDB:Display()
         oDB:UpdateSBar()
         ProcessDB( oDB )
RETURN


//
Proc testDB(cFile)
Local i
local oDb

     IF cFile == NIL .or. !file( cFile )
        ? "ERROR: NO VALID FILENAME GIVEN..."
        Quit
     endif

     USE (cFile) NEW
     oDb := DBrowseDB(2,2,20,75)
     oDb:ColorSpec := "n/w,w/b"
     oDb:headsep   := ""
     oDb:ColSep    := "  "

     for i=1 to fcount()
           oDb:AddColumn( TbColumnNew( FieldName(i), FieldBlock( FieldName(i) ) ))
     next

     oDb:Setkey( 13, {|| oDB:TbGet() })
     oDB:AddVSBar()
     oDB:AddHSBar()
     oDB:Display()
     ProcessDB( oDB )
RETURN


//
PROC ProcessDB( oDB )
LOCAL nKey

     WHILE (.T.)
       DispBegin()
       oDB:ForceStable()
       oDB:UpdateSBar()
       DispEnd()
           IF (oDB:ApplyKey((nKey := Inkey(0) ) ) == -1)
               oDB:Hide()
               EXIT
           ELSEIF nKey == K_CTRL_UP
               oDB:Move( -1, 0 )
           ELSEIF nKey == K_CTRL_LEFT
               oDB:Move( 0, -1 )
           ELSEIF nKey == K_CTRL_RIGHT
               oDB:Move( 0, 1 )
           ELSEIF nKey == K_CTRL_DOWN
               oDB:Move( 1, 0 )
           ELSEIF nKey == K_LBUTTONDOWN
               oDB:MsHandler()
               oDB:ForceStable()

           ENDIF
     ENDDO

RETURN

#endif



//------------------------------------------------------  SHELL FUNCTIONS

FUNCTION DBrowseNew(t,l,b,r)
Local o
DEFAULT t := 0, l := 0, b := MaxRow(), r := MaxCol()
o := DBrowse():TBNEW(t,l,b,r)
Return(o)


FUNCTION DBrowseDB(t,l,b,r)
Local o
DEFAULT t := 0, l := 0, b := MaxRow(), r := MaxCol()
o := DBrowse():TBDB(t,l,b,r)
Return(o)


//------------------------------------------------------  INIT CLASS

CLASS DBrowse

DATA oTB
DATA cBack
DATA cFrame    AS B_SIMPLE
DATA aArray    AS NIL
DATA nIndex    AS 1          // for array browsing
DATA oVScroll                // Vertical   ScrollBar obj
DATA oHScroll                // Horizontal ScrollBar obj
DATA slot1, slot2
DATA bInit, bEnd
DATA Buffer       AS .F.
DATA Caption
DATA HasMoved     AS .F.

METHOD autoLite(x)           INLINE GETSET( ::oTb:Autolite, x )  NOSELF
METHOD border(x)             INLINE GETSET( ::oTb:border, x )    NOSELF
METHOD cargo(x)              INLINE GETSET( ::oTb:cargo, x )     NOSELF
METHOD colCount(x)           INLINE ::oTb:colCount  NOSELF
METHOD colorSpec(x)          INLINE GETSET( ::oTb:colorSpec, x ) NOSELF
METHOD colPos(x)             INLINE GETSET( ::oTb:colPos, x )    NOSELF
METHOD colSep(x)             INLINE GETSET( ::oTb:colSep, x )    NOSELF
METHOD footSep(x)            INLINE GETSET( ::oTb:footSep, x )   NOSELF
METHOD freeze(x)             INLINE GETSET( ::oTb:freeze, x )    NOSELF
METHOD goBottomBlock(x)      INLINE GETSET( ::oTb:goBottomBlock, x ) NOSELF
METHOD goTopBlock(x)         INLINE GETSET( ::oTb:goTopBlock, x ) NOSELF
METHOD headSep(x)            INLINE GETSET( ::oTb:headSep, x )   NOSELF
METHOD hitBottom(x)          INLINE GETSET( ::oTb:hitBottom, x ) NOSELF
METHOD hitTop(x)             INLINE GETSET( ::oTb:hitTop, x )    NOSELF
METHOD leftVisible(x)        INLINE ::oTb:leftVisible  NOSELF
METHOD mColPos(x)            INLINE GETSET( ::oTb:mColPos, x)    NOSELF           // 5.3
METHOD message(x)            INLINE GETSET( ::oTb:message, x )   NOSELF           // 5.3
METHOD mRowPos(x)            INLINE GETSET( ::oTb:mRowPos, x )   NOSELF           // 5.3
METHOD nBottom(x)            INLINE GETSET( ::oTb:nBottom, x )   NOSELF
METHOD nLeft(x)              INLINE GETSET( ::oTb:nLeft, x )     NOSELF
METHOD nRight(x)             INLINE GETSET( ::oTb:nRight, x )    NOSELF
METHOD nTop(x)               INLINE GETSET( ::oTb:nTop, x )      NOSELF
METHOD rightVisible(x)       INLINE ::oTb:rightVisible NOSELF
METHOD rowCount(x)           INLINE ::oTb:rowCount     NOSELF
METHOD rowPos(x)             INLINE GETSET( ::oTb:rowPos, x )    NOSELF
METHOD skipBlock(x)          INLINE GETSET( ::oTb:skipBlock, x ) NOSELF
METHOD stable(x)             INLINE GETSET( ::oTb:stable, x )    NOSELF

METHOD down                  INLINE ::oTb:down()
METHOD end                   INLINE ::oTb:end()
METHOD goBottom              INLINE ::oTb:goBottom()
METHOD goTop                 INLINE ::oTb:goTop()
METHOD home                  INLINE ::oTb:home()
METHOD left                  INLINE ::oTb:left()
METHOD pageDown              INLINE ::oTb:pageDown()
METHOD pageUp                INLINE ::oTb:pageUp()
METHOD panEnd                INLINE ::oTb:panEnd()
METHOD panHome               INLINE ::oTb:panHome()
METHOD panLeft               INLINE ::oTb:panLeft()
METHOD panRight              INLINE ::oTb:panRight()
METHOD right                 INLINE ::oTb:right()
METHOD up                    INLINE ::oTb:up()

METHOD addColumn(a,b,c)      INLINE ::oTb:addColumn(a,b,c)
METHOD applyKey(a)           INLINE ::oTb:applyKey(a) NOSELF           // 5.3
METHOD colorRect(a,b,c,d,e)  INLINE ::oTb:colorRect(a,b,c,d,e)
METHOD colWidth(a,b)         INLINE ::oTb:colWidth(a,b)
METHOD configure()           INLINE ::oTb:configure()
METHOD deHilite()            INLINE ::oTb:deHilite()
METHOD delColumn(a,b)        INLINE ::oTb:delColumn(a,b)
METHOD forceStable()         INLINE ::oTb:forceStable()           // 5.3
METHOD getColumn(a)          INLINE ::oTb:getColumn(a) NOSELF
METHOD hilite()              INLINE ::oTb:hilite()
METHOD hitTest(a,b)          INLINE ::oTb:hitTest(a,b) NOSELF           // 5.3
METHOD insColumn(a,b)        INLINE ::oTb:insColumn(a,b)
METHOD invalidate()          INLINE ::oTb:invalidate()
METHOD refreshAll()          INLINE ::oTb:refreshAll()
METHOD refreshCurrent()      INLINE ::oTb:refreshCurrent()
METHOD setColumn(a,b)        INLINE ::oTb:setColumn(a,b)
METHOD setKey(a,b)           INLINE ::oTb:setKey(a,b)             // 5.3
METHOD setstyle(a,b)         INLINE ::oTb:setstyle(a,b)           // 5.3
METHOD stabilize()           INLINE ::oTb:stabilize()

//--- new methods ---//

METHOD NEW( t,l,b,r )
METHOD TBDB(t,l,b,r)
METHOD TBNew( t,l,b,r )
METHOD Hide()
METHOD Display()
METHOD Move( nRows, nCols )
METHOD TbGet(cClr,bReader )
METHOD AddVSbar()
METHOD AddHSbar()
METHOD UpdateSBar( l )
METHOD aAddColumn( cHead, nCol, nWidth )
METHOD MsHandler()

CLASS METHOD aSetSkip()
CLASS METHOD ATBSkiper( nCurrent, nSkip )

ENDCLASS


//------------------------------------------------------  CONSTRUCTORS

//
METHOD ::New(t,l,b,r)
//
 ::oTb:= TBrowseDB(t+1,l+1,b-1,r-1)
 ::cBack := SaveScreen(t,l,b+1,r+2)
Return( self )


//
METHOD ::TBDB(t,l,b,r)
//
 ::oTb:= TBrowseDB(t+1,l+1,b-1,r-1)
 ::cBack := SaveScreen(t,l,b+1,r+2)
Return( self )


//
METHOD ::TBNew(t,l,b,r)
//
 ::oTb:= TBrowseNew(t+1,l+1,b-1,r-1)
 ::cBack := SaveScreen(t,l,b+1,r+2)
Return( self )



//
METHOD ::aAddColumn( cHead, nCol, nWidth )
//
Local oCol
DEFAULT cHead := ""
::AddColumn( (oCol := TbColumnNEW( cHead, aBrowseBlock( self, nCol ) )) )
IF nWidth != NIL
    oCol:Width := nWidth
ENDIF
Return oCol


//
STATIC FUNCTION ABrowseBlock( oTb, nCol )
//
RETURN ( {|p| IF( PCOUNT() == 0, oTb:aArray[otb:nIndex, nCol], ;
                                 otb:aArray[otb:nIndex, nCol] := p ) } )


//
METHOD ::aSetSkip()
//
Local k := 0

  ::nIndex        := 1
  ::goTopBlock    := {|| ::nIndex := 1 }
  ::goBottomBlock := {|| ::nIndex := Len(::aArray) }

  ::skipBlock := {|x| ;
                k := IF(ABS(x) >= IF(x >= 0,;
                LEN(::aArray) - ::nIndex, ::nIndex - 1),;
                IF(x >= 0, LEN(::aArray) - ::nIndex,;
                1 - ::nIndex),;
                x), ::nIndex += k,;
                k }
RETURN self


//
METHOD ::ATBSkiper( nCurrent, nSkip )
//
   IF ( nCurrent + nSkip < 1 )

      // Would skip past the top...
      RETURN ( -nCurrent + 1 )

   ELSEIF ( nCurrent + nSkip > LEN( ::aArray ) )

      // Would skip past the bottom...
      RETURN ( LEN(::aArray) - nCurrent )

   END

   // No problem
   RETURN ( nSkip )



//
Method ::Hide()
//
Dispbegin()
     RestScreen(::nTop-1,::nLeft-1,::nBottom+2,::nRight+3, ::cBack)
     ::cBack := ""
DispEnd()
RETURN (self)


//
METHOD ::Move( nRows, nCols )
//
DEFAULT nRows := 0, nCols := 0

IF nRows <> 0 .OR. nCols <> 0
   DispBegin()
     ::Hide()
     ::nTop    += nRows
     ::nLeft   += nCols
     ::nBottom += nRows
     ::nRight  += nCols
     ::HasMoved := .T.              // Note
     ::Display()
   DispEnd()
ENDIF

Return self


//
METHOD ::AddHSBar()
//
     ::oHScroll := ScrollBar( ::nLeft+1, ::nRight-1, ::nBottom+1,,2 )
     ::oHScroll:Style     := chr(27)+""+chr(26)
     ::oHScroll:ColorSpec := WSET(_SET_SCROLLBARCOLOR)
     ::oHScroll:Current   := 1
     ::oHScroll:Total     := ::oTB:ColCount()

RETURN self



//
METHOD ::AddVSBar()
//
Local nLen := 0
IF ::aArray == NIL
    nLen := if( indexord() == 0, ;
                (select())->(Lastrec()),;
                (select())->(OrdkeyCount(OrdName())) )
ELSE
    nLen := LEN( ::aArray )
ENDIF

     ::oVScroll := ScrollBar( IF( Empty(::HeadSep), ::nTop+1,::nTop+2), ;
                              IF( Empty(::FootSep), ::nBottom, ::nBottom-1), ;
                              ::nRight+1,,1 )
     ::oVScroll:Style     := ""+""+""
     ::oVScroll:ColorSpec := WSET(_SET_SCROLLBARCOLOR)
     ::oVScroll:Current   := 1
     ::oVScroll:Total     := nLen

RETURN self


//
METHOD ::UpdateSBar(lDisplay)
//
Local nPart, nCurRec := 0

DEFAULT lDisplay := .T.

IF ::aArray != NIL
   nCurRec := ::nIndex
ELSE
   nCurRec    := IF( IndexOrd() == 0,;
                     (select())->(recno()),;
                     (select())->(ordkeyNo()) )
ENDIF

/*   -------------------   HORIZONTAL   -------------------   */
IF ::oHScroll != Nil

IF ::HasMoved
     ::oHScroll:Start := ::nLeft+1
     ::oHScroll:End   := ::nRight-1
     ::oHScroll:OffSet:= ::nBottom+1
     ::oHScroll:Total := ::oTB:Colcount
ENDIF

if ::oTb:Colpos == ::oTb:ColCount
     ::oHScroll:Current  := ::oTB:ColCount  //::oHScroll:BarLength
     ::oHScroll:ThumbPos := ::oHScroll:BarLength

elseif ::oTb:Colpos == 1
     ::oHScroll:Current  := 1
     ::oHScroll:ThumbPos := 1
Else
     nPart := Round(::oHScroll:BarLength / ::oTB:ColCount,0 )
     ::oHScroll:Current  := nPart * ::oTB:ColPos
     ::oHScroll:ThumbPos := nPart * ::oTB:ColPos
endif
     ::oHScroll:Display()
endif


/*   -------------------   VERTICAL   -------------------   */
IF ::oVScroll != Nil

IF ::HasMoved
   ::oVScroll:Start   := IF( Empty(::HeadSep), ::nTop+1,::nTop+2)
   ::oVScroll:End     := IF( Empty(::FootSep), ::nBottom, ::nBottom-1)
   ::oVScroll:OffSet  := IF( ::cFrame == NIL, ::nRight, ::nRight+1 )
ENDIF

    IF nCurRec == 1
          ::oVScroll:current  := 1
          ::oVScroll:ThumbPos := 1
     ELSEIF nCurRec == ::oVScroll:Total
          ::oVScroll:current  := ::oVScroll:Total
          ::oVScroll:thumbpos := ::oVScroll:BarLength
     ELSE
          nPart := ::oVScroll:BarLength/::oVScroll:Total
          ::oVScroll:thumbpos := round( nPart * nCurRec, 0 )
     ENDIF

     ::oVScroll:Display()

ENDIF

::HasMoved := .F.    //Note

Return( self )


//
METHOD ::Display()
//

IF ::cBack == ""
      ::cBack := SaveScreen( ::nTop-1, ::nLeft-1,::nBottom+2, ::nRight+3 )
Endif

DispBegin()

     IF ::cFrame != NIL
     DispBox( ::nTop-1, ::nLeft -1, ;
              ::nBottom+1, ::nRight+1, ;
              ::cFrame, ;
              ::oTB:Colorspec)
     ENDIF

     ::oTB:RefreshAll():Forcestable()

     ::UpdateSBar( .t. )

DispEnd()

RETURN (self)



//
METHOD ::TbGet( cClr, bReader )
//
   LOCAL nCursSave, xOldKey, xNewKey
   LOCAL column, get, nKey
   Local lRead := ReadExit(.t.)
   nCursSave   := SETCURSOR( if( ReadInsert(), 3, 1 ) )

   //::oTB:ForceStable()
   column := ::oTB:getColumn( ::oTB:colPos )
   get    := GetNew( ROW(), COL(), ;
                     column:block,;
                     column:heading ,, ;
                     IF( cClr==NIL, "w+/n", cClr) )

   get:picture   := column:picture
   get:preblock  := column:preblock
   get:postblock := column:postblock

   //IF SELECT() > 0
   //  xOldKey := IF( EMPTY(INDEXKEY()), NIL, &(INDEXKEY()) )
   //ENDIF

   // Changing to a custom reader to allow time out
   if valtype( bReader ) == "B"
          get:reader := bReader
   Endif

   ReadModal( {get} )
   SETCURSOR(nCursSave)

      ::oTB:RefreshCurrent()
      ::oTB:ForceStable()
      ReadExit(.F.)

   nKey := LASTKEY()

  /*
  IF ( nKey == K_UP .OR. nKey == K_DOWN .OR. ;
      nKey == K_PGUP .OR. nKey == K_PGDN )
      KEYBOARD CHR( nKey )
   ELSEIF ( nKey == K_ENTER )
      KEYBOARD CHR( K_RIGHT )
   ENDIF
 */
RETURN( nKey )


//
METHOD ::MsHandler( nRow, nCol )
   LOCAL nCount

      IF ( ::oTb:HitTest( nRow, nCol ) == HTCELL )

         nCount := ::oTb:mRowPos - ::oTb:RowPos
         WHILE ( nCount < 0 )
            nCount ++
            ::oTb:Up()
         ENDDO

         WHILE ( nCount > 0 )
            nCount --
            ::oTb:Down()
         ENDDO

         nCount := ::oTb:mColPos - ::oTb:ColPos
         WHILE ( nCount < 0 )
            nCount ++
            ::oTb:Left()
         ENDDO

         WHILE ( nCount > 0 )
            nCount --
            ::oTb:Right()
         ENDDO

         RETURN ( TBR_CONTINUE )


      ENDIF

   RETURN ( TBR_EXCEPTION )

