/****
*
*      oDBTest1.prg
*      oObject Demonstration.
*
*      Notes:
*      ------
*      Demonstrates use of INLINE ... NOSELF clauses
*      and inline .DBF fields assignment to class variables
*
*      Also note that the ::Recno() method behaves like an
*      ACCESS-ASSIGN method.
*
*      This class reads/writes data directly to the fields so, a ::Write()
*      is not necessary...
*/

#include "oObject.ch"

Proc Test3()
Local oDB
cls
oDB := DBF():New("addbk")
odb:AddFlds()                 // --> you can also call it in oDB:New()
oDB:Recno(5)                  // --> no chaining here !!! No self is returned
oDB:Read()
oDB:ClassInfo()
? "oDB:Name  ", oDB:Name()     // note: ()
? "oDB:Name()", oDB:Name
?
? "----------------------------------------------------"
? "First Record dump... :", oDB:Recno()
? "----------------------------------------------------"
oDB:Qout()                   // dump the ::buffer
Inkey(0)

oDB:GoBottom():Read()
? "----------------------------------------------------"
? "Last Record dump...", oDB:Recno()
? "----------------------------------------------------"
oDB:Qout()                   // dump the ::buffer
inkey(0)

oDB:Name := "Altered Name..."
? "----------------------------------------------------"
? "Last Record - oDB:Name is altered", oDB:Recno()
? "If we call oDB:Write() the record also will be altered"
? "----------------------------------------------------"
oDB:Qout()                   // dump the ::buffer
inkey(0)

oDB:Axit()
RETURN

/*---------------------------------------------------------------------------*/
CLASS DBF ALIAS DB
/*---------------------------------------------------------------------------*/
DATA Buffer     AS {}         // Must be #1 as self[1]
DATA Alias      AS ""
DATA Struc      AS {}
DATA nFlds      INLINE (::alias)->(FCount())
DATA Recno(n)   INLINE IF( n == NIL, ;
                           (::alias)->(RECNO()), ;
                           (::alias)->(dbGoTo(n)) )  ;
                           NOSELF

// Method to GET/SET a Field by position
// Used by ::AddFlds() below
DATA FldGetSet(n, xVal)  ;
                INLINE If( xVal == NIL,;
                                   (::Alias)->(FieldGet(n)), ;
                                   (::Alias)->(FieldPut(n,xVal)) ) ;
                                   NOSELF

DATA Eof()      INLINE (::alias)->(EOF()) NOSELF   // Returns .T. or .F.
DATA Bof()      INLINE (::alias)->(BOF()) NOSELF

DATA goTop()    INLINE (::alias)->(dbGoTop())
DATA goBottom() INLINE (::alias)->(dbGoBottom())

METHOD New      =   dbNew
METHOD AddFlds  =   dbAddFlds
METHOD Read     =   dbRead
METHOD Write    =   dbWrite
METHOD QOut     =   dbQOut
ENDCLASS


/*---------------------------------------------------------------------------*/
METHOD FUNCTION dbNew(cAlias)
/*---------------------------------------------------------------------------*/
USE (cAlias) NEW
::Alias := ALIAS()  //cAlias
::Struc := (::alias)->(dbStruct())
::buffer := array(  ( ::alias )->( fcount() )  )  //--> init ::buffer
::GoTop():Read()
Return(  self  )

/*---------------------------------------------------------------------------*/
METHOD FUNCTION dbRead()
/*---------------------------------------------------------------------------*/
   AEval( ::Buffer, {| u, n | ::Buffer[n] := FieldGet(n) } )
RETURN self

/*---------------------------------------------------------------------------*/
METHOD FUNCTION dbWrite()
/*---------------------------------------------------------------------------*/
   AEval( ::Buffer, {| u, n | FieldPut( n, u ) } )
RETURN self

/*---------------------------------------------------------------------------*/
METHOD FUNCTION dbQout()
/*---------------------------------------------------------------------------*/
   AEval( ::Buffer, {| e | QOut( e ) } )
RETURN self

// This method initialises INLINE IVars with fieldnames but gets
// the actual data from the ::FldGetSet() Method.
// This is a faster approach because it reads and writes directly
// on the fields value instead of the ::buffer values
/*---------------------------------------------------------------------------*/
METHOD FUNCTION dbAddFlds()
/*---------------------------------------------------------------------------*/
Local i
IF empty( ::buffer )
     ::Read()
endif

FOR i=1 TO (::alias)->(FCount())
  ::InsData( Field(i), &("{|o,xVal| o:FldGetSet("+ltrim(str(i))+", xVal) }" ))
NEXT

Return(  self  )


