/****
*       OOBJECT.CH
*
*       Header file for the oObject Class Library
*       (c) 1997-1998, Manos Aspradakis
*
*/

#ifndef _OOBJECT_CH_

#stdout :
#stdout :  oObject version 2.00
#stdout :  (c) 1997/98 Manos Aspradakis
#stdout :

/*------ IVars Scope -----------------------------------------------------*/

#define SCOPE_EXPORT        0
#define SCOPE_HIDDEN        1
#define SCOPE_VIRTUAL       2
#define SCOPE_ACCESS        3
#define SCOPE_READONLY      4
#define SCOPE_CONSTRUCTOR   5

#define SCOPE_PROTECTED SCOPE_READONLY
#define SCOPE_LOCAL     SCOPE_HIDDEN
#define SCOPE_PRIVATE   SCOPE_HIDDEN

#xcommand EXPORT:    =>   oObj:Scope := SCOPE_EXPORT
#xcommand PRIVATE:   =>   oObj:Scope := SCOPE_PRIVATE
#xcommand HIDDEN:    =>   oObj:Scope := SCOPE_HIDDEN
#xcommand LOCAL:     =>   oObj:Scope := SCOPE_LOCAL
#xcommand PROTECTED: =>   oObj:Scope := SCOPE_PROTECTED
#xcommand READONLY:  =>   oObj:Scope := SCOPE_PROTECTED
#xcommand ACCESS:    =>   oObj:Scope := SCOPE_ACCESS
#xcommand VIRTUAL:   =>   oObj:Scope := SCOPE_VIRTUAL


#define VAR  DATA
#define IVAR DATA

/*------ CLASS Creation ------------------------------------------------*/


#xCommand        CLASS <cClass> [ALIAS <aliasName>] [INHERIT <oSuper>] => ;
          CLASS <cClass> [ALIAS <aliasName>] [XFROM <oSuper>()]

#xCommand        CLASS <cClass> [ALIAS <aliasName>] [FROM <oSuper>] => ;
          CLASS <cClass> [ALIAS <aliasName>] [XFROM <oSuper>()]


#xCommand CREATE CLASS <cClass> [ALIAS <aliasName>] [FROM <oSuper>] => ;
          CLASS <cClass> [ALIAS <aliasName>] [XFROM <oSuper>()]

#xCommand CREATE CLASS <cClass> [ALIAS <aliasName>] [INHERIT <oSuper>] => ;
          CLASS <cClass> [ALIAS <aliasName>] [XFROM <oSuper>()]

#xCommand CLASS <cClass> [ALIAS <aliasName>] [ XFROM <oSuper> ] ;
          => ;
          FUNCTION <cClass>() ;;
          STATIC hClass := 0 ;;
          STATIC oObj ;;
          LOCAL oNew, cAlias:= #<aliasName> ;;
          LOCAL nPos := 0 ;;
          IF hClass  == 0 ;;
          oObj       := OBJECT():New( <"cClass">, <{oSuper}>, cAlias ) ;;
          oObj:Scope := SCOPE_EXPORT ;;
          __bAddCls( { {|| <cClass>() }, <{oSuper}> } ) ;;
          cAlias     := oObj:ClsAlias

/*-------- ADD Simple IVars ------------------------------------------------*/


#xcommand DATA <var1> [,<varN>] [AS <xValue>] ;
          => ;
          oObj:AddIVar( <"var1">,, [<xValue>], oObj:Scope,, ++nPos, cAlias ) ;
          [; oObj:AddIVar( <"varN">,, <xValue>, oObj:Scope,, ++nPos, cAlias )]
         
#xcommand DATA <var1> [AS <xValue>] ;
          => ;
          oObj:AddIVar( <"var1">,, [<xValue>], oObj:Scope,, ++nPos, cAlias )
         
#xcommand DATA <var1> [,<varN>] := <xValue> ;
          => ;
          oObj:AddIVar( <"var1">,, [<xValue>], oObj:Scope,, ++nPos, cAlias ) ;
          [; oObj:AddIVar( <"varN">,, <xValue>, oObj:Scope,, ++nPos, cAlias )]

#xcommand DATA <var1> := <xValue> ;
          => ;
          oObj:AddIVar( <"var1">,, [<xValue>], oObj:Scope,, ++nPos, cAlias )
         
         
/*--------- DATA Types -----------------------------------------------------*/


#xCommand DATA <cData> [, <cDataN>]  AS <o:INT, SHORT>  ;
          => ;
          oObj:AddIVar( <"cData">,, 0,  oObj:Scope, NIL, ++nPos, cAlias ) ;
          [; oObj:AddIVar( <"cDataN">,, 0,  oObj:Scope, NIL, ++nPos, cAlias )]


#xCommand DATA <cData> [, <cDataN>] AS FLOAT  ;
          => ;
          oObj:AddIVar( <"cData">,, 0.0 ,  oObj:Scope, NIL, ++nPos, cAlias ) ;
          [; oObj:AddIVar( <"cDataN">,, 0.0 ,  oObj:Scope, NIL, ++nPos, cAlias )]


#xCommand DATA <cData> [, <cDataN>] AS <o:CHAR, STRING>  ;
          => ;
          oObj:AddIVar( <"cData">,, "",  oObj:Scope, NIL, ++nPos, cAlias ) ;
          [; oObj:AddIVar( <"cDataN">,, "",  oObj:Scope, NIL, ++nPos, cAlias )]


#xCommand DATA <cData> [, <cDataN>] AS DATE  ;
          => ;
          oObj:AddIVar( <"cData">,, CTOD("  /  /  "), oObj:Scope, NIL, ++nPos, cAlias) ;
          [; oObj:AddIVar( <"cDataN">,, ctod("  /  /  "),  oObj:Scope, NIL, ++nPos, cAlias)]


#xCommand DATA <cData> [, <cDataN>] AS <o:ARRAY, STRUC, STRUCT> ;
          => ;
          oObj:AddIVar( <"cData">,, {}, oObj:Scope, NIL, ++nPos, cAlias  ) ;
          [; oObj:AddIVar( <"cDataN">,, {},  oObj:Scope, NIL, ++nPos, cAlias )]


#xCommand DATA <cData> [, <cDataN>] AS <o:BLOCK, CODE> ;
          => ;
          oObj:AddIVar( <"cData">,, {|| NIL },  oObj:Scope, NIL, ++nPos, cAlias  ) ;
          [; oObj:AddIVar( <"cDataN">,, {|| NIL },  oObj:Scope, NIL, ++nPos, cAlias)]


#xCommand DATA <cData> [, <cDataN>] AS <o:LOGIC, BOOL, BOOLEAN> ;
          => ;
          oObj:AddIVar( <"cData">,, .F., oObj:Scope, NIL, ++nPos, cAlias  ) ;
          [; oObj:AddIVar( <"cDataN">,, .F., oObj:Scope, NIL, ++nPos, cAlias)]


/*--------- END DATA Types -------------------------------------------------*/

/*--------- User Defined IVars ---------------------------------------------*/


#xcommand DATA <var1> [AS <xValue>] GETSET <cFunc> ;
          => ;
          oObj:AddData( <"var1">, #<cFunc>, [<xValue>], oObj:Scope,, ++nPos, cAlias )
         

/*--------- LINKed IVars ---------------------------------------------------*/


#xcommand DATA <var1> [AS <xValue>] LINK <cMethod>;
          => ;
          oObj:AddData( <"var1">,, [<xValue>], oObj:Scope, #<cMethod>, ++nPos, cAlias )



/*--------- INLINE / BLOCK IVars -------------------------------------------*/


#xCommand DATA <cData> BLOCK <bBlock> ;
          => ;
          oObj:AddData( <"cData">, <{bBlock}>,, oObj:Scope, NIL, ++nPos, cAlias )



#xCommand DATA <cData> INLINE [<xpression,...>] ;
          => ;
          oObj:AddData( <"cData">, {|Self| [<xpression>], Self },, oObj:Scope, NIL, ++nPos, cAlias )

#xCommand DATA <cData>( [<uParms,...>] ) INLINE [<xpression,...>] ;
          => ;
          oObj:AddData( <"cData">, { |Self [,<uParms>]| [<xpression>], Self },, oObj:Scope, NIL, ++nPos, cAlias )



#xCommand DATA <cData> INLINE <xpression,...> NOSELF ;
          => ;
          oObj:AddData( <"cData">, {|Self| <xpression> },, oObj:Scope, NIL, ++nPos, cAlias )
          
#xCommand DATA <cData>( [<uParms,...>] ) INLINE <xpression,...> NOSELF ;
          => ;
          oObj:AddData( <"cData">, { |Self [,<uParms>]| <xpression> },, oObj:Scope, NIL, ++nPos, cAlias )


#xcommand DATA <var1> [,<varN>] EXCLUDE => ;
          oObj:DelData( <"var1"> ) [ ; oObj:DelData( <"varN"> )]



 // @@@  Inline Methods...

#xCommand METHOD <cData> INLINE [<xpression,...>] ;
          => ;
          oObj:AddData( <"cData">, {|Self| [<xpression>], Self },, oObj:Scope, NIL, ++nPos, cAlias  )

#xCommand METHOD <cData>( [<uParms,...>] ) INLINE [<xpression,...>] ;
          => ;
          oObj:AddData( <"cData">, {|Self [,<uParms>]| [<xpression>], Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand METHOD <cData> INLINE [<xpression,...>] NOSELF;
          => ;
          oObj:AddData( <"cData">, {|Self| [<xpression>] },, oObj:Scope, NIL, ++nPos, cAlias  )

#xCommand METHOD <cData>( [<uParms,...>] ) INLINE [<xpression,...>] NOSELF;
          => ;
          oObj:AddData( <"cData">, {|Self [,<uParms>]| [<xpression>] },,;
          oObj:Scope, NIL, ++nPos, cAlias )



/*--------- On the fly ADD DATA ---------------------------------------------*/


#xCommand ADD DATA <cData> INLINE [<xpression,...>] TO <oObj> ;
          => ;
          <oObj>:InsData( <"cData">, {|Self|  [<xpression>], Self },,)

#xCommand ADD DATA <cData>( [<uParms,...>] ) INLINE [<xpression,...>] ;
          TO <oObj>;
          => ;
          <oObj>:InsData( <"cData">, { |Self [,<uParms>]|  [<xpression>], Self },,)


#xCommand ADD DATA <cData> INLINE <xpression,...> TO <oObj> ;
          NOSELF  ;
          => ;
          <oObj>:InsData( <"cData">, {|Self| <xpression> },,)

#xCommand ADD DATA <cData>( [<uParms,...>] ) INLINE <xpression,...> TO <oObj> ;
          NOSELF ;
          => ;
          <oObj>:InsData( <"cData">, { |Self [,<uParms>]| <xpression> },,)


#xcommand ADD DATA <var1>  [AS <xValue>] GETSET <cFunc> TO <oObj> ;
          => ;
          <oObj>:InsData( <"var1">, #<cFunc>, [<xValue>] )
         

#xcommand ADD DATA <var1> [AS <xValue>] LINK <cMethod> TO <oObj>;
          => ;
          <oObj>:InsData( <"var1">,, [<xValue>],, #<cMethod> )





/*----------- Add METHODS --------------------------------------------------*/

#xcommand METHOD <cMessage> [,<*cMsg*>]  [CONSTRUCTOR] ;
          => ;
          oObj:AddMethod( <"cMessage">, cAlias+<"cMessage">, cAlias, oObj:Scope ) [; METHOD <cMsg>]

#xCommand METHOD <cMessage> = <cMethod> [CONSTRUCTOR] ;
          => ;
          oObj:AddMethod( <"cMessage">, <"cMethod">, cAlias, oObj:Scope )


#xCommand MESSAGE <cMessage> METHOD <cMethod> [CONSTRUCTOR] ;
          => ;
          oObj:AddMethod( <"cMessage">, <"cMethod">, cAlias, oObj:Scope )


#xCommand METHOD <cMessage> BLOCK <bBlock> => ;
          oObj:AddData( <"cMessage">, <{bBlock}> )


#xCommand METHOD <cMessage> [= <cMethod>] VIRTUAL;
          => ;
          oObj:AddMethod( <"cMessage">, "__VIRTUAL", cAlias, oObj:Scope )


#xCommand METHOD <cMessage> [= <cMethod>] DEFERRED;
          => ;
          oObj:AddMethod( <"cMessage">, "__VIRTUAL", cAlias, oObj:Scope )


#xCommand MESSAGE <cMessage> [METHOD <cMethod>] VIRTUAL ;
          => ;
          oObj:AddMethod( <"cMessage">, "__VIRTUAL", cAlias, oObj:Scope )

#xCommand MESSAGE <cMessage> [METHOD <cMethod>] DEFERRED ;
          => ;
          oObj:AddMethod( <"cMessage">, "__VIRTUAL", cAlias, oObj:Scope )


#xcommand METHOD <cMessage> [,<*cMsg*>] EXCLUDE => ;
          oObj:DelMethod( <"cMessage"> ) [; METHOD <cMsg>]


/*--------------------- CLASS (protected) Methods ------------------------*/
              
#xcommand CLASS METHOD <cMessage> [,<*cMsg*>]  ;
          => ;
          oObj:AddMethod( <"cMessage">, cAlias+<"cMessage">, cAlias, SCOPE_LOCAL ) [; CLASS METHOD <cMsg>]

#xCommand CLASS METHOD <cMessage> , <cMethod> ;
          => ;
          oObj:AddMethod( <"cMessage">, <"cMethod">, cAlias, SCOPE_LOCAL )

#xCommand CLASS METHOD <cMessage> = <cMethod> ;
          => ;
          oObj:AddMethod( <"cMessage">, <"cMethod">, cAlias, SCOPE_LOCAL )


// ----------------- Support for the new STATIC Method UDFs ------------------

// --> Declare Methods at class init.

#xCommand METHOD <cMethod> ;
          => ;
          oObj:AddData( <"cMethod">, ;
          {|Self| <cMethod>( Self ) },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand METHOD <cMethod>( [<uParms,...>] ) ;
          => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| <cMethod>( Self [, <uParms>]) },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand METHOD <cMethod>( [<uParms,...>] ) CONSTRUCTOR ;
          => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| <cMethod>( Self [, <uParms>]) },,;
          SCOPE_CONSTRUCTOR, NIL, ++nPos, cAlias )


#xCommand CLASS METHOD <cMethod>( [<uParms,...>] ) ;
          => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| <cMethod>( Self [, <uParms>]) },,;
          SCOPE_LOCAL, NIL, ++nPos, cAlias )


// ------------------- SubClass an internal object IVar -----------------------


#xCommand DATA <cData> SUBCLASS <oIVar> => ;
          oObj:AddData( <"cData">, ;
          {|Self, xVal| ::<oIVar>:<cData>:= IF(xVal == NIL, ::<oIVar>:<cData>, xVal) },,;
          oObj:Scope, NIL, ++nPos, cAlias )


#xCommand METHOD <cMethod> SUBCLASS <oIVar> => ;
          oObj:AddData( <"cMethod">, ;
          {|Self| ::<oIVar>:<cMethod>(), Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )


#xCommand METHOD <cMethod> SUBCLASS <oIVar> NOSELF => ;
          oObj:AddData( <"cMethod">, ;
          {|Self| ::<oIVar>:<cMethod>() },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand METHOD <cMethod>( [<uParms,...>] ) SUBCLASS <oIVar> => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| ::<oIVar>:<cMethod>([<uParms>]), Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )


#xCommand METHOD <cMethod>( [<uParms,...>] ) SUBCLASS <oIVar> NOSELF => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| ::<oIVar>:<cMethod>([<uParms>]) },,;
          oObj:Scope, NIL, ++nPos, cAlias )



// The Class(y) way ...
#xCommand MESSAGE <cMethod>( [<uParms,...>] ) TO <oIVar> => ;
          oObj:AddData( <"cMethod">, ;
          {|Self [,<uParms>]| ::<oIVar>:<cMethod>([<uParms>]), Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand MESSAGE <cMethod> TO <oIVar> => ;
          oObj:AddData( <"cMethod">, ;
          {|Self| ::<oIVar>:<cMethod>, Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand MESSAGE <cMsg> IS <cMethod> IN <oIVar> => ;
          oObj:AddData( <"cMsg">, ;
          {|Self| ::<oIVar>:<cMethod>, Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )

#xCommand MESSAGE <cMsg>( [<uParms,...>] ) IS <cMethod> IN <oIVar> => ;
          oObj:AddData( <"cMsg">, ;
          {|Self [,<uParms>]| ::<oIVar>:<cMethod>([<uParms>]), Self },,;
          oObj:Scope, NIL, ++nPos, cAlias )


// ----------------------- End Class init ------------------------------------

#xCommand METHOD <cMessage> = <cMethod> [CONSTRUCTOR] ;
          => ;
          oObj:AddMethod( <"cMessage">, <"cMethod">, cAlias, oObj:Scope )

#xCommand ENDCLASS => END CLASS

#xCommand END CLASS =>                        ;
          hClass      := oObj:Insert()        ;;
          end                                 ;;
          oNew        := __ClassIns( hClass ) ;;
          oObj:NewObj := oNew                 ;;
          oNew:Dict   := oObj                 ;;
          oObj:EndInsert()                    ;;
          RETURN( oNew )                      ;;
          #include "oobject2.ch"

/*----------------------- Implement Methods ------------------------------*/

#undef METHOD

#xCommand CONSTRUCTOR <MethodName>( [<uParms,...>] ) ;
          => ;
          Function <MethodName>( [<uParms>] )  ;;
          Local Self := QSelf()


#xCommand METHOD FUNCTION <MethodName>( [<uParms,...>] ) ;
          => ;
          Function <MethodName>( [<uParms>] )  ;;
          Local Self := QSelf()


#xCommand METHOD PROCEDURE <MethodName>( [<uParms,...>] ) ;
          => ;
          PROCEDURE <MethodName>( [<uParms>] )



// --> Method function UDFs

#xTranslate METHOD FUNCTION <cMethodName>( [<uParms,...>] ) CLASS <ClassName> => ;
            STATIC FUNCTION <cMethodName>( Self [,<uParms>] )

#xTranslate METHOD <cMethodName>( [<uParms,...>] ) CLASS <ClassName> => ;
            STATIC FUNCTION <cMethodName>( Self [,<uParms>] )

#xtranslate METHOD <cClassName>::<cMethodName>( [<uParms,...>] ) => ;
            STATIC FUNCTION <cMethodName>( Self [,<uParms>] )

#xtranslate METHOD ::<cMethodName>( [<uParms,...>] ) => ;
            STATIC FUNCTION <cMethodName>( Self [,<uParms>] )

#xtranslate STATIC METHOD <cMethodName>( [<uParms,...>] ) => ;
            STATIC FUNCTION <cMethodName>( Self [,<uParms>] )



// ----------------  o:Axit PseudoMethod + Memory cleanup ---------------

#xTranslate :Axit() => := NIL ; memory(0)
#xTranslate Axit(<o>)  => <o> := NIL ; memory(0)


// -------------------  ACCESS Parent/Super -----------------------------

//#define VAR  DATA
//#define IVAR DATA

#xtranslate ::  => Self:
#xtranslate ::: => Self:dict:super:
#xTranslate ::Super => Self:Dict:Super

#xTranslate xSuper():<Msg>([<uParms,...>]) => __XSUPER(self, <"Msg">[, <uParms>])
#xTranslate ::xSuper():<Msg>([<uParms,...>]) => __XSUPER(self, <"Msg">[, <uParms>])
#xTranslate parent([<cClass>]) =>  __SUPER(#<cClass>)
#xTranslate super([<cClass>])  =>  __SUPER(#<cClass>)
#xtranslate super([<class>]):<method> => __SUPER(#<class>):<method>
#xTranslate send(<obj>, <msg>) => eval(&("{|o| o:" + #<msg> + "}"), <obj>)

#ifdef NO_DIRECT_IVARS

#xtranslate @:<msg>           => Self:nMessage(<"msg">)
#xtranslate @:<msg> := <xVal> => Self:nMessage(<"msg">, xVal)
#xtranslate @:<msg>:= <xVal>  => Self:nMessage(<"msg">, xVal)
#xtranslate @:<msg>:=<xVal>   => Self:nMessage(<"msg">, xVal)

#endif


//----------------- PSEUDO FUNCTIONS (ripped off mostly <g>) ----------------

#xtranslate BYNAME <V> [, <VN> ]     => ::<V> := <V> [; ::<VN> := <VN> ]
#xtranslate BYNAME <V> DEFAULT <Val> => ::<V> := BYDEFAULT <V>, <Val>
#xtranslate BYDEFAULT <V>, <Val>     => if( <V> == NIL, <Val>, <V> )

#xcommand DEFAULT <uVar1> := <uVal1> ;
               [, <uVarN> := <uValN> ] => ;
                  <uVar1> := If( <uVar1> == nil, <uVal1>, <uVar1> ) ;;
                [ <uVarN> := If( <uVarN> == nil, <uValN>, <uVarN> ); ]

#xcommand DEFAULT <v1> TO <x1> [, <vn> TO <xn> ]                        ;
          =>                                                            ;
          IF <v1> == NIL ; <v1> := <x1> ; END                           ;
          [; IF <vn> == NIL ; <vn> := <xn> ; END ]

#define _OOBJECT_CH_
#endif
